module.exports = {
    parse: parse
};

var Parser = require("jison").Parser;
var fs = require('fs');

eval(fs.readFileSync(__dirname + '/lib/ifc.js', 'utf8'));

function parse(stream, callback) {
    var grammar = {
        "lex": {
            "rules": [
                ["\\s+", "/* skip whitespace */"],
                //["\\/\\*[^.]*\\*\\/", "/* Skip comments */"],
                ["\\/\\*(\\*(?!\\/)|[^*])*\\*\\/", "/* Skip comments */"],
                ["=", "return 'EQUAL';"],
                ["\\(", "return 'LEFTPARENTHESIS';"],
                ["\\)", "return 'RIGHTPARENTHESIS';"],
                ["\\$", "return 'UNDEFINED';"],
                [",", "return 'COMMA';"],
                [";", "return 'SEMICOLON';"],
                ["\\*", "return 'ASTERISK';"],
                ['END-ISO-10303-21', "return 'ENDISO'"],
                ['ISO-10303-21', "return 'ISO'"],
                ['HEADER', "return 'HEADER'"],
                ['FILE_DESCRIPTION', "return 'FILEDESCRIPTION'"],
                ['FILE_NAME', "return 'FILENAME'"],
                ['FILE_SCHEMA', "return 'FILESCHEMA'"],
                ['ENDSEC', "return 'ENDSEC'"],
                ['DATA', "return 'DATA'"],
                ["#[0-9]+", "return 'OBJECTID';"],
                ["\'[^']*\'", "return 'STRING';"],
                ["\\.[^.]*\\.", "return 'KEYWORD';"],
                ["[a-zA-Z][a-zA-Z0-9]*", "return 'IDENTIFIER';"],
                ["-?[0-9]+(\\.[0-9]*)?([Ee][+-]?[0-9]+)?", "return 'NUMBER';"],
                ["$", "return 'EOF';"]
            ]
        },

        "bnf": {
            "item": [['line EOF', "return $1;"]],
            "line": [['HEADER', "return null;"], ['objectDefinition', "return $1;"], ['FILEDESCRIPTION attributeList', "return null;"], ['FILENAME attributeList', "return null;"],
                ['FILESCHEMA attributeList', "return null;"], ['ISO', "return null;"], ['ENDSEC', "return null;"], ['DATA', "return null;"], ['ENDISO', "return null;"]],
            "objectDefinition": [['OBJECTID EQUAL IDENTIFIER attributeList', "$$ = {oid: $1, type: $3, __attributes: $4}; return $$;"]],
            "attributeList": [['LEFTPARENTHESIS attributes RIGHTPARENTHESIS', '$$ = $2;'], ['LEFTPARENTHESIS RIGHTPARENTHESIS', '$$ = $2;']],
            "attributes": [['attribute COMMA attributes', '$$ = $3; $$.unshift($1);'], ['attribute', '$$ = [$1];']],
            'attribute': [['objectAttribute', '$$ = $1'], ['strings', '$$ = $1'], ['KEYWORD', '$$ = yytext.substring(1, yytext.length-1)'], ['ASTERISK', '$$ = null'], ['NUMBER', '$$ = Number(yytext)'], ['OBJECTID', '$$ = yytext'], ['UNDEFINED', '$$ = null'], ['attributeList', '$$ = $1'], ['IDENTIFIER attributeList', '$$ = $2']],
            'strings': [['string strings', '$$ = $1 + $2;'], ['string', '$$ = $1;']],
            'string': [['STRING', '$$ = yytext.substring(1, yytext.length-1);']],
            'objectAttribute': [['Identifier LEFTPARENTHESIS KEYWORD RIGHTPARENTHESIS', 'if ($1 == "IFCBOOLEAN") {$$ = ($3 == ".F." ? false : true);} else {$$ = $3;}'],
                ['Identifier LEFTPARENTHESIS String RIGHTPARENTHESIS', '$$ = $3;'], ['Identifier LEFTPARENTHESIS Number RIGHTPARENTHESIS', '$$ = $3;']]
        }
    };

    var parser = new Parser(grammar);

    //var object;
    // var line = 0;
    //var objectId;
    var line = '';
    var objects = [];
    var start;
    var end;

    start = new Date().getTime();

    stream
        .on('data', function (chunk) {
            chunk = new String(chunk);

            while (chunk.indexOf(';') >= 0) {
                line += chunk.substring(0, chunk.indexOf(';'))
                chunk = chunk.substring(chunk.indexOf(';') + 1);

                // Prevent ; in strings.
                // TODO Check for escaped '

                if (line.split('\'').length > 0 && line.split('\'').length % 2 == 0) {
                    continue;
                }

                //console.log(line.substring(0, 20) + '...' + line.substring(line.length - 20));

                var object = parser.parse(line);

                if (object) {
                    objects.push(object);
                }

                line = '';
            }

            line += chunk;
        })
        .on('end', function () {
            end = new Date().getTime();

            console.log('>>> Parsing IFC >>> ' + (end - start));

            var errors = [];
            var map = {};
            var project;

            for (var n in objects) {
                if (ifc.getPrototype(objects[n].type)) {
                    inheritMethods(objects[n], ifc.getPrototype(objects[n].type));

                    if (objects[n].type == 'IFCPROJECT') {
                        project = objects[n];
                    } else {
                        map[objects[n].oid] = objects[n];
                    }

                    objects[n].load(errors);

                    if (objects[n].__attributes) {
                        delete objects[n].__attributes;
                    }
                }
            }

            project.map = map;

            console.log('>>> Parsing Complete');

            callback(errors, project)
        });
}

function inheritMethods(target, source) {
    for (var member in source) {
        if (source[member] instanceof Function) {
            target[member] = source[member];
        }
    }
}

